# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
from __future__ import annotations

from collections.abc import Callable

from .dimensioned_array import (
    ArrayImplementation,
    DimensionedArray,
    Dims,
    UnitImplementation,
)


def _merge_dims(a: Dims, b: Dims) -> Dims:
    """Favor order in a."""
    # TODO Avoid transpose of b if possible
    return a + tuple(dim for dim in b if dim not in a)


def unary(
    x: DimensionedArray,
    values_op: Callable[[ArrayImplementation], ArrayImplementation],
    unit_op: Callable[[UnitImplementation], UnitImplementation],
) -> DimensionedArray:
    return DimensionedArray(
        values=values_op(x.values),
        dims=x.dims,
        unit=None if x.unit is None else unit_op(x.unit),
    )


def elemwise_binary(
    x,
    /,
    y: DimensionedArray,
    *,
    values_op: Callable[
        [ArrayImplementation, ArrayImplementation], ArrayImplementation
    ],
    unit_op: Callable[[UnitImplementation, UnitImplementation], UnitImplementation],
) -> DimensionedArray:
    dims = _merge_dims(x.dims, y.dims)
    a = x.values
    b = y.values
    for dim in dims:
        if dim not in x.dims:
            a = x.array_api.expand_dims(a, axis=dims.index(dim))
        if dim not in y.dims:
            b = y.array_api.expand_dims(b, axis=0)
    b_dims = (*(set(dims) - set(y.dims)), *y.dims)
    b = y.array_api.permute_dims(b, axes=tuple(dims.index(dim) for dim in b_dims))
    return DimensionedArray(
        values=values_op(a, b),
        dims=dims,
        unit=(
            None
            # TODO do not mix unit with None
            if x.unit is None and y.unit is None
            else unit_op(x.unit, y.unit)
        ),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
import numpy as np
import pytest

import pydims as dms
from pydims.testing import assert_identical


def test_concat():
    da = dms.DimensionedArray(values=np.ones((2, 3)), dims=('x', 'y'), unit=None)
    assert_identical(
        dms.concat((da, da), dim='x'),
        dms.DimensionedArray(values=np.ones((4, 3)), dims=('x', 'y'), unit=None),
    )
    assert_identical(
        dms.concat((da, da), dim='y'),
        dms.DimensionedArray(values=np.ones((2, 6)), dims=('x', 'y'), unit=None),
    )


def test_reshape_raises_NotImplementedError():
    with pytest.raises(
        NotImplementedError, match="`reshape` is deliberately not supported"
    ):
        dms.reshape()


def test_stack():
    da = dms.DimensionedArray(values=np.ones((4, 3)), dims=('x', 'y'), unit=None)
    assert_identical(
        dms.stack((da, da), dim='z'),
        dms.DimensionedArray(
            values=np.ones((2, 4, 3)), dims=('z', 'x', 'y'), unit=None
        ),
    )
    assert_identical(
        dms.stack((da, da), dim='z', axis=0),
        dms.DimensionedArray(
            values=np.ones((2, 4, 3)), dims=('z', 'x', 'y'), unit=None
        ),
    )
    assert_identical(
        dms.stack((da, da), dim='z', axis=1),
        dms.DimensionedArray(
            values=np.ones((4, 2, 3)), dims=('x', 'z', 'y'), unit=None
        ),
    )
    assert_identical(
        dms.stack((da, da), dim='z', axis=2),
        dms.DimensionedArray(
            values=np.ones((4, 3, 2)), dims=('x', 'y', 'z'), unit=None
        ),
    )
    assert_identical(
        dms.stack((da, da), dim='z', axis=-1),
        dms.DimensionedArray(
            values=np.ones((4, 3, 2)), dims=('x', 'y', 'z'), unit=None
        ),
    )
    assert_identical(
        dms.stack((da, da), dim='z', axis=-2),
        dms.DimensionedArray(
            values=np.ones((4, 2, 3)), dims=('x', 'z', 'y'), unit=None
        ),
    )
    assert_identical(
        dms.stack((da, da), dim='z', axis=-3),
        dms.DimensionedArray(
            values=np.ones((2, 4, 3)), dims=('z', 'x', 'y'), unit=None
        ),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)

import scipp as _scipp

dimensionless = _scipp.units.dimensionless
Unit = _scipp.units.Unit


def get_scale(*, src, dst) -> float:
    q_src = _scipp.scalar(1, unit=src)
    q_dst = _scipp.to_unit(q_src, unit=dst)
    return float(q_dst.value)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
import array_api_strict as array
from scipp import units

import pydims as dms
from pydims.testing import assert_identical

make = dms.CreationFunctions(array, units.Unit)


def test_create_with_scipp_unit():
    x = make.linspace('x', 0, 1, 3, unit='m')
    assert x.unit == units.Unit('m')


def test_to_unit():
    x = make.linspace('x', 0, 1, 3, unit='m')
    y = x.to(unit='cm')
    assert y.unit == units.Unit('cm')
    assert_identical(y, make.linspace('x', 0, 100, 3, unit='cm'))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Pydims contributors (https://github.com/pydims)
# ruff: noqa: E402, F401

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

from .creation_functions import CreationFunctions
from .dimensioned_array import DimensionedArray, DimensionError, exp
from .array_api_manipulation_functions import concat, reshape, stack
from .array_api_utility_functions import all, any

__all__ = [
    'all',
    'any',
    'CreationFunctions',
    'DimensionedArray',
    'DimensionError',
    'exp',
    'concat',
    'reshape',
    'stack',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
"""
Functions from the "Utility Functions" section of the Python Array API.

"""

from .dimensioned_array import Dim, DimensionedArray, Dims


def all(x: DimensionedArray, /, *, dim: Dim | Dims | None = None) -> DimensionedArray:
    """
    Tests whether all input array elements evaluate to `True` along a given dimension.

    Parameters
    ----------
    x:
        Input array
    dim:
        Dimension or dimensions along which to perform a logical AND reduction.

    Returns
    -------
    :
        Result array
    """
    if x.unit is not None:
        # TODO Is this what we want? Is there any harm in allowing a unit?
        raise ValueError("Unit is not supported for logical operation")
    axis, dims = _axis_dims_for_reduce(x, dim)
    # boolean output has unit=None, input unit is discarded
    return DimensionedArray(
        values=x.array_api.all(x.values, axis=axis), dims=dims, unit=None
    )


def any(x: DimensionedArray, /, *, dim: Dim | Dims | None = None) -> DimensionedArray:
    """
    Tests whether any input array elements evaluate to `True` along a given dimension.

    Parameters
    ----------
    x:
        Input array
    dim:
        Dimension or dimensions along which to perform a logical OR reduction.

    Returns
    -------
    :
        Result array
    """
    if x.unit is not None:
        # TODO Is this what we want? Is there any harm in allowing a unit?
        raise ValueError("Unit is not supported for logical operation")
    axis, dims = _axis_dims_for_reduce(x, dim)
    # boolean output has unit=None, input unit is discarded
    return DimensionedArray(
        values=x.array_api.any(x.values, axis=axis), dims=dims, unit=None
    )


def _axis_dims_for_reduce(x, dim):
    dim = dim or x.dims
    if not isinstance(dim, tuple):
        dim = (dim,)

    axis = tuple(x.dims.index(d) for d in dim)
    dims = tuple(d for d in x.dims if d not in dim)
    return axis, dims


__all__ = ['all', 'any']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
"""
Functions from the "Manipulation Functions" section of the Python Array API.

Will not support:

- broadcast_arrays
- broadcast_to
- reshape -> use fold and flatten instead
"""

import array_api_compat

from .dimensioned_array import Dim, DimensionedArray


def concat(
    arrays: tuple[DimensionedArray, ...], /, *, dim: Dim | None = None
) -> DimensionedArray:
    """
    Concatenate arrays along a given dimension.

    Parameters
    ----------
    arrays:
        Arrays to concatenate.
    dim:
        Dimension along which to concatenate. If None, arrays must be 1-D.

    Returns
    -------
    :
        Concatenated array.
    """
    first = arrays[0]
    dim = dim or first.dim
    if not all(arr.dims == first.dims for arr in arrays):
        raise ValueError("All arrays must have the same dims")
    if not all(arr.unit == first.unit for arr in arrays):
        raise ValueError("All arrays must have the same unit")
    axis = first.dims.index(dim)
    values = [arr.values for arr in arrays]
    xp = array_api_compat.array_namespace(*values)
    return DimensionedArray(
        values=xp.concat(values, axis=axis), dims=first.dims, unit=first.unit
    )


def reshape(*args, **kwargs):
    raise NotImplementedError(
        "`reshape` is deliberately not supported `fold` and `flatten` instead"
    )


def stack(
    arrays: tuple[DimensionedArray, ...] | list[DimensionedArray],
    /,
    *,
    dim: Dim,
    axis: int = 0,
) -> DimensionedArray:
    """
    Stack arrays along a new dimension.

    Parameters
    ----------
    arrays:
        Arrays to stack.
    dim:
        Dimension along which to stack.
    axis:
        Location of the new dimension.

    Returns
    -------
    :
        Stacked array.
    """
    first = arrays[0]
    if dim in first.dims:
        raise ValueError("Dimension already exists, did you mean to use `concat`?")
    if not all(arr.dims == first.dims for arr in arrays):
        raise ValueError("All arrays must have the same dims")
    if not all(arr.unit == first.unit for arr in arrays):
        raise ValueError("All arrays must have the same unit")
    dims = list(first.dims)
    dims.insert(axis if axis >= 0 else first.ndim + 1 + axis, dim)
    values = [arr.values for arr in arrays]
    xp = array_api_compat.array_namespace(*values)
    return DimensionedArray(
        values=xp.stack(values, axis=axis), dims=dims, unit=first.unit
    )


__all__ = ['concat', 'reshape', 'stack']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
import array_api_strict as array
import astropy.units as u

import pydims as dms
from pydims.testing import assert_identical

make = dms.CreationFunctions(array, u.Unit)


def test_from_astropy_units():
    x = make.linspace('x', 0, 1, 3, unit='m')
    assert_identical(
        x,
        dms.DimensionedArray(values=array.linspace(0, 1, 3), dims=('x',), unit=u.meter),
    )
    assert (x + x).unit == u.meter
    assert (x * x).unit == u.meter**2


def test_unit_conversion():
    m = make.linspace('x', 0, 1, 3, unit='m')
    km = m.to(unit='km')
    assert km.unit == u.km
    assert all(km.values == m.values / 1000)


def test_units_namespace_detects_astropy():
    from pydims.units_api import units_namespace

    units = units_namespace(u.meter)
    assert units is not None
    assert units.Unit('m') == u.meter



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
from __future__ import annotations

from pydims.string_unit import StringUnit as _Unit

dimensionless = _Unit()

Unit = _Unit



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)

import pydims as dms


def assert_identical(a: dms.DimensionedArray, b: dms.DimensionedArray):
    # TODO This is a quick hack for early testing
    assert a.dims == b.dims  # noqa: S101
    assert a.shape == b.shape  # noqa: S101
    assert a.dtype == b.dtype  # noqa: S101
    assert a.unit == b.unit  # noqa: S101
    assert a.array_api.all(a.values == b.values)  # noqa: S101



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
from __future__ import annotations

import astropy.units as _units

dimensionless = _units.dimensionless_unscaled

Unit = _units.Unit


def get_scale(*, src, dst) -> float:
    return src.to(dst)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class StringUnit:
    """Very simple unit class that stores a unit as a string."""

    value: str = ''

    def __mul__(self, other: StringUnit) -> StringUnit:
        return StringUnit(f'({self.value})*({other.value})')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
import array_api_strict as array
import pytest

import pydims as dms
from pydims.string_unit import StringUnit
from pydims.testing import assert_identical


@pytest.mark.parametrize('dims', [(), ('x',), ('x', 'y', 'z')])
def test_init_raises_if_dims_has_wrong_length(dims: tuple[str, ...]):
    with pytest.raises(ValueError, match="Number of dimensions"):
        dms.DimensionedArray(values=array.ones((2, 3)), dims=dims, unit=None)


def test_sizes():
    da = dms.DimensionedArray(values=array.ones((2, 3)), dims=('x', 'y'), unit=None)
    assert da.sizes == {'x': 2, 'y': 3}


def test_getitem_no_dim_raises_if_not_1d():
    da = dms.DimensionedArray(values=array.ones((2, 3)), dims=('x', 'y'), unit=None)
    with pytest.raises(
        dms.DimensionError, match="Only 1-D arrays can be indexed without dims"
    ):
        _ = da[0]


def test_getitem_1d_no_dim():
    da = dms.DimensionedArray(dims=('x',), values=array.arange(2.0), unit=None)
    assert_identical(
        da[0], dms.DimensionedArray(dims=(), values=array.asarray(0.0), unit=None)
    )
    assert_identical(
        da[1], dms.DimensionedArray(dims=(), values=array.asarray(1.0), unit=None)
    )


def test_getitem_2d_with_dims():
    da = dms.DimensionedArray(
        values=array.reshape(array.arange(6), (2, 3)), dims=('x', 'y'), unit=None
    )
    assert_identical(
        da[{'x': 0}],
        dms.DimensionedArray(dims=('y',), values=array.asarray([0, 1, 2]), unit=None),
    )
    assert_identical(
        da[{'y': 0}],
        dms.DimensionedArray(dims=('x',), values=array.asarray([0, 3]), unit=None),
    )
    assert_identical(
        da[{'x': 1, 'y': 1}],
        dms.DimensionedArray(dims=(), values=array.asarray(4), unit=None),
    )


@pytest.mark.parametrize('unit', [None, StringUnit(), StringUnit('m')])
def test_neg(unit: StringUnit | None):
    da = dms.DimensionedArray(values=array.ones((2, 3)), dims=('x', 'y'), unit=unit)
    result = -da
    assert_identical(
        result,
        dms.DimensionedArray(values=-array.ones((2, 3)), dims=('x', 'y'), unit=unit),
    )


def test_exp_raises_if_unit_is_not_dimensionless():
    da = dms.DimensionedArray(
        values=array.ones((2, 3)), dims=('x', 'y'), unit=StringUnit('m')
    )
    with pytest.raises(ValueError, match="Unit must be dimensionless"):
        dms.exp(da)


def test_add_raises_if_units_differ():
    da1 = dms.DimensionedArray(
        values=array.ones((2, 3)), dims=('x', 'y'), unit=StringUnit('m')
    )
    da2 = dms.DimensionedArray(
        values=array.ones((2, 3)), dims=('x', 'y'), unit=StringUnit('s')
    )
    with pytest.raises(ValueError, match="Units must be identical"):
        da1 + da2


def test_elemwise_binary_broadcasts_dims():
    xy = dms.DimensionedArray(values=array.ones((2, 3)), dims=('x', 'y'), unit=None)
    yz = dms.DimensionedArray(values=array.ones((3, 4)), dims=('y', 'z'), unit=None)
    result = dms.common.elemwise_binary(
        xy, yz, values_op=lambda a, b: a + b, unit_op=lambda a, b: a
    )
    assert result.dims == ('x', 'y', 'z')


def test_elemwise_binary_transposes_dims():
    xy = dms.DimensionedArray(values=array.ones((2, 3)), dims=('x', 'y'), unit=None)
    yx = dms.DimensionedArray(values=array.ones((3, 2)), dims=('y', 'x'), unit=None)
    result = dms.common.elemwise_binary(
        xy, yx, values_op=lambda a, b: a + b, unit_op=lambda a, b: a
    )
    assert result.dims == ('x', 'y')


def test_elemwise_binary_broadcasts_and_transposes_dims():
    xy = dms.DimensionedArray(values=array.ones((2, 3)), dims=('x', 'y'), unit=None)
    yxz = dms.DimensionedArray(
        values=array.ones((3, 2, 4)), dims=('y', 'x', 'z'), unit=None
    )
    result = dms.common.elemwise_binary(
        xy, yxz, values_op=lambda a, b: a + b, unit_op=lambda a, b: a
    )
    assert result.sizes == {'x': 2, 'y': 3, 'z': 4}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
from __future__ import annotations

import sys
from typing import Any


def _is_astropy_unit(unit: Any) -> bool:
    if 'astropy.units' not in sys.modules:
        return False
    import astropy.units

    return issubclass(unit.__class__, astropy.units.UnitBase)


def _is_pint_unit(unit: Any) -> bool:
    if 'pint' not in sys.modules:
        return False
    import pint

    return issubclass(unit.__class__, pint.Unit)


def is_scipp_unit(unit: Any) -> bool:
    if 'scipp' not in sys.modules:
        return False
    import scipp

    return isinstance(unit, scipp.units.Unit)


def _is_string_unit(unit: Any) -> bool:
    from pydims.string_unit import StringUnit

    return issubclass(unit.__class__, StringUnit)


def units_namespace(unit: Any) -> Any:
    if _is_astropy_unit(unit):
        from . import astropy

        return astropy
    elif _is_pint_unit(unit):
        from .pint import PintsUnitsNamespace

        return PintsUnitsNamespace(unit._REGISTRY)
    elif is_scipp_unit(unit):
        from . import scipp

        return scipp
    elif _is_string_unit(unit):
        from . import string_unit

        return string_unit


__all__ = ['units_namespace']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Pydims contributors (https://github.com/pydims)
# ruff: noqa: E402, F401

from .units_api import units_namespace

__all__ = ['units_namespace']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
import numpy as np

import pydims as dms
from pydims.testing import assert_identical


def test_all():
    da = dms.DimensionedArray(
        values=np.array([[True, True, False], [True, True, True]]),
        dims=('x', 'y'),
        unit=None,
    )
    assert_identical(
        dms.all(da, dim='x'),
        dms.DimensionedArray(
            values=np.array([True, True, False]), dims=('y',), unit=None
        ),
    )
    assert_identical(
        dms.all(da, dim='y'),
        dms.DimensionedArray(values=np.array([False, True]), dims=('x',), unit=None),
    )


def test_any():
    da = dms.DimensionedArray(
        values=np.array([[False, True, False], [False, False, False]]),
        dims=('x', 'y'),
        unit=None,
    )
    assert_identical(
        dms.any(da, dim='x'),
        dms.DimensionedArray(
            values=np.array([False, True, False]), dims=('y',), unit=None
        ),
    )
    assert_identical(
        dms.any(da, dim='y'),
        dms.DimensionedArray(values=np.array([True, False]), dims=('x',), unit=None),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
import dask.array as da
import numpy as np

import pydims as dms
from pydims.common import unary
from pydims.testing import assert_identical


def compute(x):
    return x.compute()


def test_create_from_chunked_dask_array():
    values = da.ones((10, 10), chunks=(5, 5))
    dims = ('x', 'y')
    a = dms.DimensionedArray(values=values, dims=dims, unit=None)
    b = a + a
    # Binary op should not trigger computation
    assert not isinstance(b.values, np.ndarray)
    result = unary(b, values_op=lambda x: x.compute(), unit_op=None)
    assert isinstance(result.values, np.ndarray)
    assert_identical(
        result, dms.DimensionedArray(values=2 * np.ones((10, 10)), dims=dims, unit=None)
    )


def test_slice_chunked_dask_array():
    values = da.ones((10, 10), chunks=(5, 5))
    dims = ('x', 'y')
    a = dms.DimensionedArray(values=values, dims=dims, unit=None)
    b = a[{'x': slice(4, 8), 'y': slice(6, 8)}]
    # Slicing should not compute
    assert not isinstance(b.values, np.ndarray)
    result = unary(b, values_op=lambda x: x.compute(), unit_op=None)
    assert isinstance(result.values, np.ndarray)
    assert_identical(
        result, dms.DimensionedArray(values=np.ones((4, 2)), dims=dims, unit=None)
    )


def test_can_create_chunked_using_CreationFunctions():
    make = dms.CreationFunctions(da, None)
    x = make.linspace('x', 0, 1, 4, unit=None, chunks=(2,))
    assert x.values.chunks == ((2, 2),)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
from __future__ import annotations

from pint import UnitRegistry


class PintsUnitsNamespace:
    def __init__(self, ureg: UnitRegistry):
        self._ureg = ureg

    def __getattr__(self, item):
        return getattr(self._ureg, item)

    def get_scale(self, *, src, dst) -> float:
        q_src = self._ureg.Quantity(1, src)
        q_dst = q_src.to(dst)
        return q_dst.magnitude



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
from __future__ import annotations

import operator
from collections.abc import Hashable, Iterator, Mapping
from typing import Any, Protocol

import array_api_compat

from . import units_api

DType = Any  # Is the array API standard defining a DType type?


class DimensionError(Exception):
    pass


class ArrayImplementation(Protocol):
    """Array of values following the Python array API standard."""


class UnitImplementation(Protocol):
    pass


# Tuple is hashable so this leads to some ambiguity, but if in doubt a tuple is
# interpreted as a tuple of dimensions, not a single dimension.
Dim = Hashable
Dims = tuple[Dim, ...]
Shape = tuple[int, ...]


class Sizes(Mapping):
    def __init__(self, dims: Dims, shape: Shape):
        self._data = dict(zip(dims, shape, strict=True))

    def __getitem__(self, key) -> int:
        return self._data[key]

    def __iter__(self) -> Iterator[Dim]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)


class DimensionedArray:
    """
    Array with named dimensions and optional unit.
    """

    def __init__(
        self,
        *,
        values: ArrayImplementation,
        dims: Dims,
        unit: UnitImplementation | None,
    ):
        if len(dims) != values.ndim:
            raise ValueError(
                f"Number of dimensions ({values.ndim}) does "
                f"not match number of dims ({len(dims)})"
            )
        self._values = values
        self._dims = tuple(dims)
        self._unit = unit

    def __str__(self) -> str:
        return (
            f"dims={self.dims}\nshape={self.shape}\n"
            f"values={self.values}\nunit={self.unit}"
        )

    def _repr_html_(self) -> str:
        return (
            f"<table>"
            f"<tr><td>dims</td><td>{self.dims}</td></tr>"
            f"<tr><td>shape</td><td>{self.shape}</td></tr>"
            f"<tr><td>values</td><td>{self.values}</td></tr>"
            f"<tr><td>unit</td><td>{self.unit}</td></tr>"
            f"</table>"
        )

    @property
    def array_api(self) -> Any:
        return array_api_compat.array_namespace(self.values)

    @property
    def unit_api(self) -> Any:
        return units_api.units_namespace(self.unit)

    @property
    def dtype(self) -> DType:
        return self._values.dtype

    @property
    def ndim(self) -> int:
        return self._values.ndim

    @property
    def size(self) -> int:
        return self._values.size

    @property
    def dims(self) -> Dims:
        return self._dims

    @property
    def dim(self) -> Dim:
        if len(self.dims) != 1:
            raise DimensionError("Number of dimensions must be 1")
        return self.dims[0]

    @property
    def unit(self) -> UnitImplementation | None:
        return self._unit

    @property
    def shape(self) -> Shape:
        return tuple(self._values.shape)

    @property
    def sizes(self) -> Sizes:
        return Sizes(dims=self.dims, shape=self.shape)

    @property
    def values(self) -> ArrayImplementation:
        return self._values

    def astype(self, dtype: DType, copy: bool = True) -> DimensionedArray:
        return DimensionedArray(
            values=self.array_api.astype(self.values, dtype, copy=copy),
            dims=self.dims,
            unit=self.unit,
        )

    def _to_unit(self, unit: Any, copy: bool = True) -> DimensionedArray:
        scale = self.unit_api.get_scale(src=self.unit, dst=unit)
        if scale == 1 and not copy:
            return self
        return DimensionedArray(
            values=self.values * scale,
            dims=self.dims,
            unit=self.unit_api.Unit(unit),
        )

    def to(
        self, *, dtype: DType | None = None, unit: Any | None = None, copy: bool = True
    ) -> DimensionedArray:
        """
        Convert to a new dtype and/or unit.

        Parameters
        ----------
        dtype:
            New dtype, None if no conversion is needed.
        unit:
            New unit, None if no conversion is needed.
        copy:
            If True, a copy of the values is made, even if no conversion is needed.

        Returns
        -------
        :
            New array with the requested dtype and/or unit.
        """
        if unit is None and dtype is None:
            raise ValueError("Must provide dtype or unit or both")

        if dtype is None:
            return self._to_unit(unit, copy=copy)

        if unit is None:
            return self.astype(dtype, copy=copy)

        # TODO Logic copied from Scipp. Probably not complete with the Array API dtypes
        api = self.array_api
        if dtype == api.float64:
            convert_dtype_first = True
        elif self.dtype == api.float64:
            convert_dtype_first = False
        elif dtype == api.float32:
            convert_dtype_first = True
        elif self.dtype == api.float32:
            convert_dtype_first = False
        elif self.dtype == api.int64 and dtype == api.int32:
            convert_dtype_first = False
        elif self.dtype == api.int32 and dtype == api.int64:
            convert_dtype_first = True
        else:
            convert_dtype_first = True

        if convert_dtype_first:
            return self.to(dtype=dtype, copy=copy).to(unit=unit, copy=False)
        else:
            return self.to(unit=unit, copy=copy).to(dtype=dtype, copy=False)

    def __getitem__(
        self, key: int | slice | dict[Dim, int | slice]
    ) -> DimensionedArray:
        if isinstance(key, int | slice):
            if self.ndim != 1:
                raise DimensionError("Only 1-D arrays can be indexed without dims")
            key = {self.dim: key}

        dims = tuple(dim for dim in self.dims if not isinstance(key.get(dim), int))
        values_key = tuple(key.pop(dim, slice(None)) for dim in self.dims)
        if key:
            raise ValueError(f"Unknown dimensions: {key.keys()}")
        return DimensionedArray(
            values=self.values[values_key], dims=dims, unit=self.unit
        )

    def __neg__(self) -> DimensionedArray:
        from .common import unary

        return unary(
            self, values_op=self.values.__class__.__neg__, unit_op=_unchanged_unit
        )

    def __add__(self, other: DimensionedArray) -> DimensionedArray:
        from .common import elemwise_binary

        return elemwise_binary(
            self,
            other,
            values_op=self.values.__class__.__add__,
            unit_op=_same_unit,
        )

    def __mul__(self, other: DimensionedArray) -> DimensionedArray:
        from .common import elemwise_binary

        return elemwise_binary(
            self,
            other,
            values_op=self.values.__class__.__mul__,
            unit_op=operator.mul,
        )


def _unchanged_unit(unit: UnitImplementation) -> UnitImplementation:
    return unit


def _same_unit(a: UnitImplementation, b: UnitImplementation) -> UnitImplementation:
    if a != b:
        raise ValueError("Units must be identical")
    return a


def _unit_must_be_dimensionless(unit: UnitImplementation) -> UnitImplementation:
    if unit * unit != unit:
        raise ValueError("Unit must be dimensionless")
    return unit


def exp(x: DimensionedArray, /) -> DimensionedArray:
    from .common import unary

    return unary(x, values_op=x.array_api.exp, unit_op=_unit_must_be_dimensionless)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
from __future__ import annotations

from typing import Any

from array_api_compat import array_namespace

from .dimensioned_array import (
    ArrayImplementation,
    Dim,
    DimensionedArray,
    Dims,
    Shape,
    UnitImplementation,
)
from .units_api import units_namespace

_default_unit = object()


class CreationFunctions:
    def __init__(self, array_api: Any, unit_api: Any):
        # TODO What should we pass here? Modules? A Unit class?
        self._array_api = array_namespace(array_api.zeros(0))
        self._unit_api = None if unit_api is None else units_namespace(unit_api(''))

    def _maybe_unit(
        self, unit: Any | UnitImplementation | None, values: ArrayImplementation
    ) -> UnitImplementation | None:
        if unit is _default_unit:
            if self._array_api.isdtype(values.dtype, 'numeric'):
                return self._unit_api.dimensionless
            else:
                return None
        return None if unit is None else self._unit_api.Unit(unit)

    def arange(
        self,
        dim: Dim,
        start: float,
        /,
        stop: float,
        step: float = 1,
        *,
        unit: Any | UnitImplementation | None = _default_unit,
        **kwargs: Any,  # dtype, device
    ) -> DimensionedArray:
        values = self._array_api.arange(start, stop, step, **kwargs)
        return DimensionedArray(
            values=values, dims=(dim,), unit=self._maybe_unit(unit, values)
        )

    def asarray(
        self,
        dims: Dims,
        values: Any,
        *,
        unit: Any | UnitImplementation | None = _default_unit,
        **kwargs: Any,  # dtype, device, copy
    ) -> DimensionedArray:
        values = self._array_api.asarray(values, **kwargs)
        return DimensionedArray(
            values=values, dims=dims, unit=self._maybe_unit(unit, values)
        )

    def linspace(
        self,
        dim: Dim,
        start: complex,
        stop: complex,
        /,
        num: int,
        *,
        unit: Any | UnitImplementation | None = _default_unit,
        **kwargs: Any,  # dtype, device, endpoint
    ) -> DimensionedArray:
        values = self._array_api.linspace(start, stop, num, **kwargs)
        return DimensionedArray(
            values=values, dims=(dim,), unit=self._maybe_unit(unit, values)
        )

    def empty(
        self,
        dims: Dims,
        shape: Shape,
        *,
        unit: Any | UnitImplementation | None = _default_unit,
        **kwargs: Any,  # dtype, device
    ) -> DimensionedArray:
        values = (self._array_api.empty(shape, **kwargs),)
        return DimensionedArray(
            values=values, dims=dims, unit=self._maybe_unit(unit, values)
        )

    def ones(
        self,
        dims: Dims,
        shape: Shape,
        *,
        unit: Any | UnitImplementation | None = _default_unit,
        **kwargs: Any,  # dtype, device
    ) -> DimensionedArray:
        values = self._array_api.ones(shape, **kwargs)
        return DimensionedArray(
            values=values, dims=dims, unit=self._maybe_unit(unit, values)
        )

    def zeros(
        self,
        dims: Dims,
        shape: Shape,
        *,
        unit: Any | UnitImplementation | None = _default_unit,
        **kwargs: Any,  # dtype, device
    ) -> DimensionedArray:
        values = self._array_api.zeros(shape, **kwargs)
        return DimensionedArray(
            values=values, dims=dims, unit=self._maybe_unit(unit, values)
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 PyDims contributors (https://github.com/pydims)
import numpy as np

import pydims as dms
from pydims.string_unit import StringUnit
from pydims.testing import assert_identical


def test_create_creation_functions_from_numpy():
    make = dms.CreationFunctions(np, StringUnit)
    x = make.linspace('x', 0, 1, 3, unit='m')
    assert_identical(
        x,
        dms.DimensionedArray(
            values=np.linspace(0, 1, 3), dims=('x',), unit=StringUnit('m')
        ),
    )


def test_arange():
    make = dms.CreationFunctions(np, StringUnit)
    x = make.arange('x', 0, 1, 0.5, unit='m')
    assert_identical(
        x,
        dms.DimensionedArray(
            values=np.arange(0, 1, 0.5), dims=('x',), unit=StringUnit('m')
        ),
    )


def test_default_unit_depends_on_dtype():
    make = dms.CreationFunctions(np, StringUnit)
    assert make.zeros(dims=('x',), shape=(2,)).unit == StringUnit()
    assert make.zeros(dims=('x',), shape=(2,), dtype=bool).unit is None
